/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file IPbasic_Errors.h
 * Header definition of the static class Errors.
 *
 * @author A. Huber
 * @since 26.10.2016
 */

#ifndef ERRORS_H
#define ERRORS_H
#include <ostream>
#include <string>
#include <vector>

#include "../IPbasic.h"
#include "IPbasic_Myfunction.h"

namespace IPbasic
{

    class Errors
    {
      public:
        /// Definition of the struct type ErrorDataPoint.
        struct ErrorDataPoint
        {
            /// Counter of the error.
            int_type counter;
            /// ID or Indes of the error..
            int_type Index;
            /// Errormessage.
            string_type MyError;

            /**
             * \brief DataPoint:
             *        Default constructor of the struct type ErrorDataPoint.
            \*/
            ErrorDataPoint()
                : counter(0)
                , Index(0)
                , MyError("")
            {
            }

            /**
             * \brief operator<<:
             *        Streamoperator of the struct type ErrorDataPoint.
             *
             * \param[in] out   Used stream.
             * \param[in] a     Printed ErrorDataPoint.
             *
             * @return Used stream.
             *
            \*/
            friend ostream_type & operator<<(ostream_type & out, ErrorDataPoint const & a)
            {
                out << a.Index << '\t' << a.counter << '\t' << a.MyError;
                return out;
            }
        };

        /// Definition of the struct type WarningDataPoint.
        struct WarningDataPoint
        {
            /// Counter of the warning.
            int_type counter;
            /// ID or Indes of the warning.
            int_type Index;
            /// Warningmessage.
            string_type MyWarning;

            /**
             * \brief DataPoint:
             *        Default constructor of the struct type WarningDataPoint.
            \*/
            WarningDataPoint()
                : counter(0)
                , Index(0)
                , MyWarning("")
            {
            }

            /**
             * \brief operator<<:
             *        Streamoperator of the struct type WarningDataPoint.
             *
             * \param[in] out   Used stream.
             * \param[in] a     Printed WarningDataPoint.
             *
             * @return Used stream.
             *
            \*/
            friend ostream_type & operator<<(ostream_type & out, WarningDataPoint const & a)
            {
                out << a.Index << '\t' << a.counter << '\t' << a.MyWarning;
                return out;
            }
        };

        /**
         * \brief set_Error:
         *        Pushes an error into this static class.
         *
         * \param[in] Error     Errormessage.
         *
        \*/
        static void set_Error(string_type const & Error);

        /**
         * \brief set_Warning:
         *        Pushes an warning into this static class.
         *
         * \param[in] Warning     Warningmessage.
         *
        \*/
        static void set_Warning(string_type const & Warning);

        /**
         * \brief get_Error_Counter:
         *        Get the number of errors occurred.
         *
         * @return The number of errors occurred.
         *
        \*/
        static int_type get_Error_Counter();

        /**
         * \brief get_Warning_Counter:
         *        Get the number of warnings occurred.
         *
         * @return The number of warnings occurred.
         *
        \*/
        static int_type get_Warning_Counter();

        /**
         * \brief get_akt_Error_string:
         *        Get the last errormessage.
         *
         * @return The last errormessage.
         *
        \*/
        static string_type get_akt_Error_string();

        /**
         * \brief get_akt_Warning_string:
         *        Get the last warningmessage.
         *
         * @return The last warningmessage.
         *
        \*/
        static string_type get_akt_Warning_string();

        /**
         * \brief get_all_Errors:
         *        Get all errormessages in a std::vector.
         *
         * @return Returns all errormessages in a std::vector.
         *
        \*/
        static std::vector<ErrorDataPoint> get_all_Errors();

        /**
         * \brief get_all_Warnings:
         *        Get all warningmessages in a std::vector.
         *
         * @return Returns all warningmessages in a std::vector.
         *
        \*/
        static std::vector<WarningDataPoint> get_all_Warnings();

        /**
         * \brief printErrorsToFile:
         *        Prints errors and warnings to a file.
         *
         * \param[in] FileName  Filename with path.
         *
        \*/
        static void printErrorsToFile(string_type const & FileName);

        /**
         * \brief resetErrors:
         *        Resets all errors.
        \*/
        static void resetErrors();

        /**
         * \brief resetWarnings:
         *        Resets all warnings.
        \*/
        static void resetWarnings();

        /**
         * \brief setExternalErrorFunction:
         *        Sets the external error function.
        \*/
        static void setExternalErrorFunction(void (*function)(string_type const &));

        /**
         * \brief setExternalWarningFunction:
         *        Sets the external warning function.
        \*/
        static void setExternalWarningFunction(void (*function)(string_type const &));

        /// Are the errors printed to the console (Default: True).
        static bool PrintErrors;

        /// Are the warnings printed to the console (Default: True).
        static bool PrintWarnings;

        /// Are the errors printed colored (Default: True).
        static bool ColoredErrors;

        /// Are the warnings printed colored (Default: True).
        static bool ColoredWarnings;

        /// Are the errors thrown as exceptions (Default: False).
        static bool ThrowErrors;

      private:
        /**
         *  External error function.
        \*/
        static MyFunction externalError;

        /**
         *  External warning function.
        \*/
        static MyFunction externalWarning;

        /// Last errormessage.
        static string_type Akt_Error;
        /// Counter of the errors.
        static int_type Error_counter;
        /// Vector with all errors.
        static std::vector<ErrorDataPoint> MemErrors;

        /**
         * \brief PushError:
         *        Errorsmessage is pushed into the error pool.
         *
         * \param[in] Error     Errormessage.
         *
        \*/
        static void PushError(string_type const & Error);

        /// Last warningmessage.
        static string_type Akt_Warning;
        /// Counter of the warnings.
        static int_type Warning_counter;
        /// Vector with all warnings.
        static std::vector<WarningDataPoint> MemWarning;

        /**
         * \brief PushWarning:
         *        Warningmessage is pushed into the error pool.
         *
         * \param[in] Error     Warningmessage.
         *
        \*/
        static void PushWarning(string_type const & Error);
    };

} // namespace IPbasic

#endif // ERRORS_H
