/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/** \mainpage IPbasic:
 *
 * This library allows solving nonlinear optimization problems (nlps) by an interior-point-method.
 *
 * The nonlinear optimization problems are characterized by the formulation
 * \f[
 * \min f(x) \qquad f : \mathbb{R}^{n_{\mathrm{state}}} \rightarrow \mathbb{R}
 * \f]
 * with the constraints
 * \f[
 * c(x) \leq 0 \qquad c : \mathbb{R}^{n_{\mathrm{state}}} \rightarrow \mathbb{R}^{n_{\mathrm{c}}}.
 * \f]
 *
 * The constraints \f$c(x)\f$  may include equality constraints and inequality constraints
 * \f[
 * h(x) = 0, \qquad g(x) \leq 0.
 * \f]
 *
 * More information on the nonlinear optimization problem can be found in the IPbasic::NLP class.
 *
 *      Copyright (C) 2019:
 *
 *            Enrico Bertolazzi (1) & Andreas Huber (2)
 *            (1) Dipartimento di Ingegneria Industriale
 *                Universita` degli Studi di Trento
 *                email: enrico.bertolazzi@unitn.it
 *            (2) Fakultät für Luft- und Raumfahrttechnik
 *                Universität der Bundeswehr München
 *                email: info@ocpbasic.com
 */

///
/// file: IPbasic.h
///

#ifndef IPBASIC_H
#define IPBASIC_H

#include "IPbasic_Config.h"

#include <lapack_wrapper/TicToc.hh>
#include <lapack_wrapper/lapack_wrapper++.hh>
#include <lapack_wrapper/lapack_wrapper.hh>

// select computer architecture
#ifdef LAPACK_WRAPPER_OS_OSX
#define IPBASIC_OS_OSX 1
#endif

#ifdef LAPACK_WRAPPER_OS_LINUX
#define IPBASIC_OS_LINUX 1
#endif

#ifdef LAPACK_WRAPPER_OS_WINDOWS
#define IPBASIC_OS_WINDOWS 1
#endif

#ifdef LAPACK_WRAPPER_ARCH32
#define IPBASIC_ARCH32 1
#endif

#ifdef LAPACK_WRAPPER_ARCH64
#define IPBASIC_ARCH64 1
#endif

// check if compiler is C++11
#ifdef LAPACK_WRAPPER_USE_CXX11
#define IPBASIC_USE_CXX11
#else
namespace std
{
    template <class InputIt, class Size, class OutputIt>
    OutputIt copy_n(InputIt first, Size count, OutputIt result)
    {
        return std::copy(first, first + count, result);
    }
} // namespace std
#endif

#define IPBASIC_PURE_VIRTUAL LAPACK_WRAPPER_PURE_VIRTUAL
#define IPBASIC_OVERRIDE LAPACK_WRAPPER_OVERRIDE

#ifndef IPBASIC_ASSERT
#define IPBASIC_ASSERT(COND, MSG)                                                                                                                                                                      \
    if (!(COND))                                                                                                                                                                                       \
    {                                                                                                                                                                                                  \
        std::ostringstream ost;                                                                                                                                                                        \
        ost << "On line: " << __LINE__ << " file: " << __FILE__ << '\n' << MSG << '\n';                                                                                                                \
        throw std::runtime_error(ost.str());                                                                                                                                                           \
    }
#endif

#ifndef IPBASIC_WARNING
#define IPBASIC_WARNING(MSG) std::cout << "Warning: " << MSG << '\n'
#endif

namespace IPbasic
{
    typedef lapack_wrapper::doublereal real_type;
    typedef lapack_wrapper::integer    int_type;
    typedef std::string                string_type;
    typedef std::vector<bool>          vec_bool;

    using lapack_wrapper::ostream_type;

    using lapack_wrapper::m_1_pi;
    using lapack_wrapper::m_1_sqrt2;
    using lapack_wrapper::m_2_pi;
    using lapack_wrapper::m_2_sqrtpi;
    using lapack_wrapper::m_2pi;
    using lapack_wrapper::m_e;
    using lapack_wrapper::m_pi;
    using lapack_wrapper::m_pi_2;
    using lapack_wrapper::m_pi_4;
    using lapack_wrapper::m_sqrt2;
    using lapack_wrapper::m_sqrtpi;

    using lapack_wrapper::isInfinite;
    using lapack_wrapper::isNaN;
    using lapack_wrapper::isRegular;
    using lapack_wrapper::isZero;

    using lapack_wrapper::DiagMatrixWrapper;
    using lapack_wrapper::Malloc;
    using lapack_wrapper::MatrixWrapper;
    using lapack_wrapper::SparseCCOOR;
    using lapack_wrapper::SparseMatrixBase;

    using lapack_wrapper::absmax;

    extern real_type const machepsi;
    extern real_type const machepsi10;
    extern real_type const machepsi100;
    extern real_type const machepsi1000;
    extern real_type const sqrtMachepsi;

} // namespace IPbasic

#endif

///
/// eof: IPbasic.h
///
