/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file IPbasic_IPdata.h
 * Header definitions of the class IPdata.
 *
 * @author A. Huber
 * @since 05.02.2019
\*/

#ifndef IPBASIC_IPDATA_H
#define IPBASIC_IPDATA_H

#include "Errors/mycolordef.h"
#include "IPbasic.h"
#include "IPbasic_NLP.h"
#include "IPbasic_Parameter.h"
#include "IPbasic_Timings.h"

#include <vector>

namespace IPbasic
{

    template <typename real>
    class IterationDataPoint;

    template <typename real>
    ostream_type & operator<<(ostream_type & out, IterationDataPoint<real> const & a);

    /// Definition of the class IterationDataPoint.
    template <typename real>
    class IterationDataPoint
    {
      public:
        /// Counter of the iteration.
        int_type counter;
        /// Current barrier parameter \f$\mu\f$.
        real mu;
        /// Current stepsize \f$\alpha\f$.
        real alpha;
        /// Current stepsize \f$\alpha_s\f$.
        real alpha_s;
        /// Current stepsize \f$\alpha_l\f$ of the lagrange multipliers.
        real alpha_l;
        /// Value of the target function.
        real obj;
        /// Error in the KKT-conditions of the barrier problem.
        real Delta_KKT;
        /// Scaled error in the KKT-conditions of the barrier problem.
        real scaled_Delta_KKT;
        /// Size of the search direction.
        real Delta_d;
        /// Difference in the target function.
        real Delta_obj;
        /// Error in the constraints.
        real Delta_CSTR;

      private:
        /// Tolerace of the barrier parameter \f$\mu\f$.
        real tolMu;
        /// Tolerace of the KKT-error.
        real tolKKT;
        /// Tolerace of the stepsize \f$\alpha\f$.
        real tolAlpha;
        /// Printlevel.
        Printlevel print;

      public:
        /*!
         * \brief IterationDataPoint:
         *        Default constructor of the class IterationDataPoint.
         */
        IterationDataPoint(real _tolMu = 0.0, real _tolKKT = 0.0, real _tolAlpha = 0.0, Printlevel _print = PRINT_NONE)
            : counter(0)
            , mu(0.0)
            , alpha(0.0)
            , alpha_s(0.0)
            , alpha_l(0.0)
            , obj(0.0)
            , Delta_KKT(0.0)
            , scaled_Delta_KKT(0.0)
            , Delta_d(0.0)
            , Delta_obj(0.0)
            , Delta_CSTR(0.0)
            , tolMu(_tolMu)
            , tolKKT(_tolKKT)
            , tolAlpha(_tolAlpha)
            , print(_print)
        {
        }

        IterationDataPoint(const IterationDataPoint & copy)
        {
            this->counter          = copy.counter;
            this->mu               = copy.mu;
            this->alpha            = copy.alpha;
            this->alpha_s          = copy.alpha_s;
            this->alpha_l          = copy.alpha_l;
            this->obj              = copy.obj;
            this->Delta_KKT        = copy.Delta_KKT;
            this->scaled_Delta_KKT = copy.scaled_Delta_KKT;
            this->Delta_d          = copy.Delta_d;
            this->Delta_obj        = copy.Delta_obj;
            this->Delta_CSTR       = copy.Delta_CSTR;
            this->tolMu            = copy.tolMu;
            this->tolKKT           = copy.tolKKT;
            this->tolAlpha         = copy.tolAlpha;
            this->print            = copy.print;
        }

        /**
         * \brief ~IterationDataPoint:
         *        Virtual destructor of the class IterationDataPoint.
         */
        virtual ~IterationDataPoint()
        {
        }

        IterationDataPoint const & operator=(IterationDataPoint const & rhs)
        {
            this->counter          = rhs.counter;
            this->mu               = rhs.mu;
            this->alpha            = rhs.alpha;
            this->alpha_s          = rhs.alpha_s;
            this->alpha_l          = rhs.alpha_l;
            this->obj              = rhs.obj;
            this->Delta_KKT        = rhs.Delta_KKT;
            this->scaled_Delta_KKT = rhs.scaled_Delta_KKT;
            this->Delta_d          = rhs.Delta_d;
            this->Delta_obj        = rhs.Delta_obj;
            this->Delta_CSTR       = rhs.Delta_CSTR;
            this->tolMu            = rhs.tolMu;
            this->tolKKT           = rhs.tolKKT;
            this->tolAlpha         = rhs.tolAlpha;
            this->print            = rhs.print;
            return *this;
        }

        /*!
         * \brief check_INF_NAN:
         *        Checks IterationDataPoint for inf and nan.
         *
         * \return True if anything in the IterationDataPoint is inf and nan.
         *
         */
        bool check_INF_NAN() const
        {
            bool is_regular = // isRegular( this->counter )          &&
                isRegular(this->mu) && isRegular(this->alpha) && isRegular(this->alpha_s) && isRegular(this->alpha_l) && isRegular(this->obj) && isRegular(this->Delta_KKT) &&
                isRegular(this->scaled_Delta_KKT) && isRegular(this->Delta_d) && isRegular(this->Delta_obj) && isRegular(this->Delta_CSTR);
            return !is_regular;
        }

      public:
        void printHeader(ostream_type & out);
        void printLine(ostream_type & out);

        /*!
         * \brief operator<<:
         *        Streamoperator of the class IterationDataPoint.
         *
         * \param[in] out   Used stream.
         * \param[in] a     Printed IterationDataPoint.
         *
         * \return Used stream.
         *
         */
        friend ostream_type & operator<<<>(ostream_type & out, IterationDataPoint<real> const & a);
    };

    template <typename real>
    class IPdata
    {
        Malloc<real> mem_real;

      public:
        /*!
         * \brief IPdata:
         *        Constructor of the class IPdata.
         *
         * \param[in] Problem       NLP-Problem, which has to be solved.
         * \param[in] IPParams      Interior-point-parameters.
         * \param[in] Timings       Timings of the interior-point-method.
         *
         */
        IPdata(NLP<real> * Problem, IPparameter<real> * IPParams = nullptr, IPtimings<real> * Timings = nullptr);

        /*!
         * \brief ~IPdata:
         *        Virtual destructor of the class IPdata.
         */
        virtual ~IPdata();

        /*\
        :|:             _    _ _                      _
        :|:   _ __ _  _| |__| (_)__   _ __  ___ _ __ | |__  ___ _ _
        :|:  | '_ \ || | '_ \ | / _| | '  \/ -_) '  \| '_ \/ -_) '_|
        :|:  | .__/\_,_|_.__/_|_\__| |_|_|_\___|_|_|_|_.__/\___|_|
        :|:  |_|
        \*/

        /*!
         * \brief initDataFromProblem:
         *        Initializes this IPdata form the Problem.
         */
        void initDataFromProblem();

        /*!
         * \brief getProblem:
         *        Returns the nlp-problem.
         *
         * \return Returns the nlp-problem.
         *
         */
        NLP<real> * getProblem() const;

        /*!
         * \brief getTimings:
         *        Returns the IPtimings.
         *
         * \return Returns the IPtimings.
         *
         */
        IPtimings<real> * getTimings() const;

        /*!
         * \brief getIPparameter:
         *        Retruns the parameter of the interior-point-method.
         *
         * \return Retruns the parameter of the interior-point-method.
         *
         */
        IPparameter<real> * getIPparameter() const;

        /*!
         * \brief checkForValidMembers:
         *        Checks if the membervariables are set.
         *
         * \return True if the membervariables are set. Else false.
         *
         */
        bool checkForValidMembers() const;

        /*!
         * \brief updateData:
         *        Makes an update of the IP data (Hessian, Jacobian, etc.).
         */
        void updateData();

        /*!
         * \brief get_Delta_d:
         *        Get scaling factor of the scaled lagrange.
         *
         * \return Scaling factor of the scaled lagrange.
         *
         */
        virtual real get_Delta_d() const;

        /*!
         * \brief get_Delta_c:
         *        Get scaling factor of \f$-\mu e + S \lambda = 0\f$.
         *
         * \return Scaling factor of \f$-\mu e + S \lambda = 0\f$.
         *
         */
        virtual real get_Delta_c() const;

        /*!
         * \brief check_INF_NAN:
         *        Checks if there are inf or nan in the IPdata.
         *
         * \return True, if there are inf or nan in the IPdata.
         *
         */
        bool check_INF_NAN() const;

        /*!
         * \brief getTau:
         *        Get current value of \f$\tau\f$.
         *
         * \param[in] index   Index of the constraint.
         *
         * @return Current value of \f$\tau\f$.
         *
         */
        real getTau(int_type index) const;

        /*!
         * \brief updateMuSequence:
         *        Update mu sequence.
         */
        void updateMuSequence() const;

        /*\
        :|:             _    _ _                   _      _    _
        :|:   _ __ _  _| |__| (_)__  __ ____ _ _ _(_)__ _| |__| |___ ___
        :|:  | '_ \ || | '_ \ | / _| \ V / _` | '_| / _` | '_ \ / -_|_-<
        :|:  | .__/\_,_|_.__/_|_\__|  \_/\__,_|_| |_\__,_|_.__/_\___/__/
        :|:  |_|
        \*/

        /// Scaling of the target function (currently 1.0).
        real sigma;
        /// Dimension of the vector with the current states.
        int_type DimStates;
        /// Vector with the current states.
        real * states;
        /// Dimension of the vector with the Lagrange multipliers.
        int_type DimLambda;
        /// Vector with the lagrange multipliers.
        real * lambda;
        /// Vector with the slack variables.
        real * slack;
        /// Vector which contails information about eq or in_eq.
        vec_bool is_equality_vec;
        /// Vector of the Constraints.
        real * ConstraintsVector;
        /// Vector with the gradient of the target function.
        real * grad_objective;
        /// Pointer to the constraint_jacobian:
        SparseMatrixBase<real> * constraint_jacobian;
        /// Pointer to the lagrangian_hessian:
        SparseMatrixBase<real> * lagrangian_hessian;

        // Data of the interior-point-method.
        /// Current iteration data:
        IterationDataPoint<real> IterData;

        // Search directions:
        /// Searchdirection for the states.
        real * IP_d_states;
        /// Searchdirection for the lagrange multipliers.
        real * IP_d_lambda;
        /// Searchdirection for the slack variables.
        real * IP_d_slack;

        /// Tells wether there is a initial guess for the slack variables.
        bool has_initial_guess_slack;
        /// Tells wether there is a initial guess for the lagrange multipliers.
        bool has_initial_guess_lambda;
        /// Sequence of mu.
        real * mu_sequ;
        /// Maximal \f$ \mu \f$ in mu_sequ.
        real mu_max;

      protected:
        /*\
        :|:                _          _          _                _      _    _
        :|:   _ __ _ _ ___| |_ ___ __| |_ ___ __| | __ ____ _ _ _(_)__ _| |__| |___ ___
        :|:  | '_ \ '_/ _ \  _/ -_) _|  _/ -_) _` | \ V / _` | '_| / _` | '_ \ / -_|_-<
        :|:  | .__/_| \___/\__\___\__|\__\___\__,_|  \_/\__,_|_| |_\__,_|_.__/_\___/__/
        :|:  |_|
        \*/

        /// NLP-Problem, which has to be solved.
        NLP<real> * Problem;
        /// Parameter of the interior-point-method.
        IPparameter<real> * Parameter;
        /// Timings of the interior-point-method.
        IPtimings<real> * Timings;
        /// True, if the parameters were allocated internal.
        bool allocedParameters;
        /// True, if the timings were allocated internal.
        bool allocedTimings;
        /// Local Timer for data class.
        TicToc LocalDataTimer;
    };

} // namespace IPbasic

#endif // IPBASIC_IPDATA_H
