/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file IPbasic_KKTSystem.h
 * Header definitions of the class KKTSystem.
 *
 * @author A. Huber
 * @since 05.02.2019
\*/

#ifndef IPBASIC_KKTSYSTEM_H
#define IPBASIC_KKTSYSTEM_H

#include "IPbasic.h"
#include "IPbasic_Data.h"
#include "LinSolvers/IPbasic_LinearSolver_interface.h"

namespace IPbasic
{

    template <typename real>
    class KKTSystem
    {
        Malloc<real> mem_real;

      public:
        /*!
         * \brief KKTSystem:
         *        Constructor of the class KKTSystem.
         *
         * \param[in] IPD           Data of the interior-point-method.
         * \param[in] LinSolver     Solver of the linear system.
         *
         */
        KKTSystem(IPdata<real> const * IPD, LinearSolver_interface<real> * LinSolver);

        /*!
         * \brief ~KKTSystem:
         *        Virtual destructor of the class KKTSystem.
         */
        virtual ~KKTSystem();

        /*\
        :|:             _    _ _                      _
        :|:   _ __ _  _| |__| (_)__   _ __  ___ _ __ | |__  ___ _ _ ___
        :|:  | '_ \ || | '_ \ | / _| | '  \/ -_) '  \| '_ \/ -_) '_(_-<
        :|:  | .__/\_,_|_.__/_|_\__| |_|_|_\___|_|_|_|_.__/\___|_| /__/
        :|:  |_|
        \*/

        void setPointerIPMatrix(SparseMatrixBase<real> * _IPmatrix)
        {
            this->IPmatrix = _IPmatrix;
        }

        /*!
         * \brief updateSearchDirection:
         *        Makes an Update of the searchdirection.
         *
         * \param[in,out] IPD   Data of the interior-point-method.
         *
         * \return True if the searchdirection could be evaluated.
         *
         */
        bool updateSearchDirection(IPdata<real> const * IPD);

        /*!
         * \brief getKKTErrors:
         *        Gets the current KKTErrors.
         *
         * \param[in]  IPD              Data of the interior-point-method.
         * \param[out] DelKKT           Error of the KKT-conditions.
         * \param[out] ScaledDelKKT     Scaled error of the KKT-conditions.
         * \param[out] DelConstr        Error of the constraints.
         *
         */
        void getKKTErrors(IPdata<real> const * IPD, real & DelKKT, real & ScaledDelKKT, real & DelConstr);

        /*!
         * \brief CreateInitialGuessLambda:
         *        Creates an initial guess for the lagrange multipliers.
         *
         * \param[in]  IPD              Data of the interior-point-method.
         *
         * \return True if initial guess could be created.
         *
         */
        bool CreateInitialGuessLambda(IPdata<real> const * IPD);

        /*!
         * \brief checkForValidMembers:
         *        Checks if the membervariables are set.
         *
         * @return True if the membervariables are set. Else false.
         *
        \*/
        bool checkForValidMembers();

      protected:
        /*\
        :|:                _          _          _                _      _    _
        :|:   _ __ _ _ ___| |_ ___ __| |_ ___ __| | __ ____ _ _ _(_)__ _| |__| |___ ___
        :|:  | '_ \ '_/ _ \  _/ -_) _|  _/ -_) _` | \ V / _` | '_| / _` | '_ \ / -_|_-<
        :|:  | .__/_| \___/\__\___\__|\__\___\__,_|  \_/\__,_|_| |_\__,_|_.__/_\___/__/
        :|:  |_|
        \*/

        /// Solver of the linear system.
        LinearSolver_interface<real> * LinSolver;
        /// Newton matrix of the interior-point-method.
        SparseMatrixBase<real> * IPmatrix;
        /// Right-hand-side or (-1)*KKT-vector.
        real * RHS;
        /// KKT-Vector.
        real * KKT;
        /// Local timer inside the KKT class.
        TicToc LocalTimerKKT;
        /// Local timer inside the KKT class for the IPSystem.
        TicToc LocalTimerIPSystem;

        /*\
        :|:                _          _          _                   _
        :|:   _ __ _ _ ___| |_ ___ __| |_ ___ __| |  _ __  ___ _ __ | |__  ___ _ _ ___
        :|:  | '_ \ '_/ _ \  _/ -_) _|  _/ -_) _` | | '  \/ -_) '  \| '_ \/ -_) '_(_-<
        :|:  | .__/_| \___/\__\___\__|\__\___\__,_| |_|_|_\___|_|_|_|_.__/\___|_| /__/
        :|:  |_|
        \*/

        /*!
         * \brief getKKTVector:
         *        Returns the current KKT-vector.
         *
         * \param[in]  IPD              Data of the interior-point-method.
         * \param[out] KKT              KKT-vector.
         *
         */
        void getKKTVector(IPdata<real> const * IPD, real KKT[]);

        /*!
         * \brief getRHSVector:
         *        Returns the current RHS-vector.
         *
         * \param[in]  IPD              Data of the interior-point-method.
         * \param[out] RHS              RHS-vector.
         *
         */
        void getRHSVector(IPdata<real> const * IPD, real RHS[]);

        /*!
         * \brief getIPMatrix:
         *        Returns the newton matrix of the interior-point-method.
         *
         * \param[in]  IPD              Data of the interior-point-method.
         * \param[out] IP_Sparse        Newton matrix of the interior-point-method.
         *
         */
        void getIPMatrix(IPdata<real> const * IPD, SparseMatrixBase<real> * IP_Sparse);

        /*!
         * \brief getInitialGuessLambdaRHS:
         *        Get the Right-hand-side to solve the lambda initial guess.
         *
         * \param[in] IPD           Data of the interior-point-method.
         * \param[in] IGLambdaRHS   Right-hand-side to solve the lambda initial guess.
         *
         */
        void getInitialGuessLambdaRHS(IPdata<real> const * IPD, real * IGLambdaRHS);

        /*!
         * \brief getInitialGuessLambdaMatrix:
         *        Get the Matrix to solve the lambda initial guess.
         *
         * \param[in] IPD           Data of the interior-point-method.
         * \param[in] IGLambda      Matrix to solve the lambda initial guess.
         *
         */
        void getInitialGuessLambdaMatrix(IPdata<real> const * IPD, SparseMatrixBase<real> * IGLambda);
    };

} // namespace IPbasic

#endif // IPBASIC_KKTSYSTEM_H
