/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file IPbasic_NLP.h
 * Header definitions of the class IPbasic::NLP.
 *
 * @author A. Huber
 * @since 05.02.2019
 */

/*!
 * \class IPbasic::NLP
 *
 * The nonlinear optimization problems are characterized by the formulation
 * \f[
 * \min f(x) \qquad f : \mathbb{R}^{n_{\mathrm{state}}} \rightarrow \mathbb{R}
 * \f]
 * with the constraints
 * \f[
 * c(x) \leq 0 \qquad c : \mathbb{R}^{n_{\mathrm{state}}} \rightarrow \mathbb{R}^{n_{\mathrm{c}}}.
 * \f]
 *
 * The constraints \f$c(x)\f$  may include equality constraints and inequality constraints
 * \f[
 * h(x) = 0, \qquad g(x) \leq 0.
 * \f]
 *
 * The condition, if a constraint is an equality or an inequality constraint is characterized by a bool vector
 * \f[
 * b \in \{0,1\}^{n_{\mathrm{state}}},
 * \f]
 * where \f$1\f$ is an equality constraint and \f$0\f$ is an inequality constraint.
 *
 */

#ifndef IPBASIC_NLP_H
#define IPBASIC_NLP_H

#include "IPbasic.h"

namespace IPbasic
{

    /*\
    :|:   ___          _      ___         _    _
    :|:  | _ ) __ _ __(_)__  | _ \_ _ ___| |__| |___ _ __
    :|:  | _ \/ _` (_-< / _| |  _/ '_/ _ \ '_ \ / -_) '  \
    :|:  |___/\__,_/__/_\__| |_| |_| \___/_.__/_\___|_|_|_|
    \*/

    template <typename real>
    class NLP
    {

        typedef SparseMatrixBase<real> SPMAT;

        bool InternalAllocatedIPMatrix;

      protected:
        SPMAT * constraint_jacobian;
        SPMAT * lagrangian_hessian;
        /// Matrix used by the interior point method:
        /// Please make shure it can contain the matrix dimension
        /// (dimStates*dimLambda)x(dimStates*dimLambda)
        SPMAT * IPMatrix;
        bool *  is_equality_vec;
        /// when true, only the lower part of the hessian is stored
        bool stored_symmetric_lagrangian_hessian;
        /// when true, only the lower part of the IP matrix is stored
        bool stored_symmetric_IPMatrix;

      public:
        NLP()
            : constraint_jacobian(nullptr)
            , lagrangian_hessian(nullptr)
            , IPMatrix(nullptr)
            , is_equality_vec(nullptr)
            , stored_symmetric_lagrangian_hessian(true)
            , stored_symmetric_IPMatrix(true)
        {
            InternalAllocatedIPMatrix = false;
        }

        NLP(SPMAT * cj, SPMAT * lh, bool * iseq, bool lagrangian_hessian_is_symmetric = true)
            : constraint_jacobian(cj)
            , lagrangian_hessian(lh)
            , IPMatrix(nullptr)
            , is_equality_vec(iseq)
            , stored_symmetric_lagrangian_hessian(lagrangian_hessian_is_symmetric)
            , stored_symmetric_IPMatrix(true)
        {
            InternalAllocatedIPMatrix = false;
        }

        NLP(SPMAT * cj, SPMAT * lh, SPMAT * ip, bool * iseq, bool lagrangian_hessian_is_symmetric = true, bool IPmatrix_is_symmetric = true)
            : constraint_jacobian(cj)
            , lagrangian_hessian(lh)
            , IPMatrix(ip)
            , is_equality_vec(iseq)
            , stored_symmetric_lagrangian_hessian(lagrangian_hessian_is_symmetric)
            , stored_symmetric_IPMatrix(IPmatrix_is_symmetric)
        {
            InternalAllocatedIPMatrix = false;
        }

        virtual ~NLP()
        {
            if (this->InternalAllocatedIPMatrix)
            {
                delete this->IPMatrix;
                this->IPMatrix = nullptr;
            }
        }

        bool checkValidMembers()
        {
            if (nullptr == this->IPMatrix && !this->InternalAllocatedIPMatrix)
            {
                this->IPMatrix = new SparseCCOOR<real>();
                int_type N     = this->state_dimension() + this->constraint_dimension();
                int_type nnz   = this->constraint_jacobian->get_nnz() + this->lagrangian_hessian->get_nnz() + this->constraint_dimension();
                this->IPMatrix->init(N, N, nnz, false);
                this->InternalAllocatedIPMatrix = true;
            }
            else if (nullptr == this->IPMatrix && this->InternalAllocatedIPMatrix)
            {
                return false;
            }

            if (nullptr == this->constraint_jacobian)
                return false;
            if (nullptr == this->lagrangian_hessian)
                return false;
            if (nullptr == this->lagrangian_hessian)
                return false;
            if (nullptr == this->is_equality_vec)
                return false;
            return true;
        }

        virtual void initial_guess(real x[]) const IPBASIC_PURE_VIRTUAL;

        virtual bool initial_guess_slack(real /*slack*/[]) const
        {
            return false;
        }

        virtual bool initial_guess_lambda(real /*lambda*/[]) const
        {
            return false;
        }

        virtual int_type state_dimension() const IPBASIC_PURE_VIRTUAL;

        virtual int_type constraint_dimension() const IPBASIC_PURE_VIRTUAL;

        virtual int_type number_of_equality_constraints() const;

        virtual int_type number_of_unilateral_constraints() const;

        bool is_equality(int_type k) const
        {
            return is_equality_vec[k];
        }

        bool is_lagrangian_hessian_symmetric() const
        {
            return this->stored_symmetric_lagrangian_hessian;
        }

        bool is_IPmatrix_symmetric() const
        {
            return this->stored_symmetric_IPMatrix;
        }

        /*\
         |   ___  _     _        _   _
         |  / _ \| |__ (_)___ __| |_(_)_ _____
         | | (_) | '_ \| / -_) _|  _| \ V / -_)
         |  \___/|_.__// \___\__|\__|_|\_/\___|
         |           |__/
        \*/

        virtual real objective(real const x[]) IPBASIC_PURE_VIRTUAL;

        virtual void objective_gradient(real const x[], real grad[]) IPBASIC_PURE_VIRTUAL;

        /*\
         |    ___             _            _     _
         |   / __|___ _ _  __| |_ _ _ __ _(_)_ _| |_ ___
         |  | (__/ _ \ ' \(_-<  _| '_/ _` | | ' \  _(_-<
         |   \___\___/_||_/__/\__|_| \__,_|_|_||_\__/__/
         |
        \*/

        virtual void constraint(real const x[], real c[]) IPBASIC_PURE_VIRTUAL;

        virtual void constraint_jacobian_update(real const x[]) IPBASIC_PURE_VIRTUAL;

        /*\
         |   _                                 _
         |  | |   __ _ __ _ _ _ __ _ _ _  __ _(_)__ _ _ _
         |  | |__/ _` / _` | '_/ _` | ' \/ _` | / _` | ' \
         |  |____\__,_\__, |_| \__,_|_||_\__, |_\__,_|_||_|
         |            |___/              |___/
        \*/

        virtual void lagrangian_hessian_update(real sigma, real const x[], real const lambda[]) IPBASIC_PURE_VIRTUAL;

        SPMAT * get_constraint_jacobian() const
        {
            return this->constraint_jacobian;
        }
        SPMAT * get_lagrangian_hessian() const
        {
            return this->lagrangian_hessian;
        }
        SPMAT * get_IPMatrix() const
        {
            return this->IPMatrix;
        }
    };

    template <typename real>
    inline int_type NLP<real>::number_of_equality_constraints() const
    {
        int_type n  = 0;
        int_type nc = this->constraint_dimension();
        for (int_type k = 0; k < nc; ++k)
            if (this->is_equality_vec[k])
                ++n;
        return n;
    }

    template <typename real>
    inline int_type NLP<real>::number_of_unilateral_constraints() const
    {
        return this->constraint_dimension() - this->number_of_equality_constraints();
    }

} // namespace IPbasic

#endif

///
/// eof: IPbasic.h
///
