/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file IPbasic_Parameter.h
 * Header definitions of the class IPparameter.
 *
 * @author A. Huber
 * @since 05.02.2019
 */

#ifndef IPBASIC_PARAMETER_H
#define IPBASIC_PARAMETER_H

#include "IPbasic.h"

namespace IPbasic
{

    /**
     * \brief Flags for the definition of Printlevels.
     */
    typedef enum
    {
        /// No printing.
        PRINT_NONE = 0,
        /// Printing only solution.
        PRINT_LESS = 1,
        /// Printing iterations and solution.
        PRINT_MORE = 2,
        /// Printing iterations and solution with colors.
        PRINT_MARKED = 3
    } Printlevel;

    /**
     * \brief Flags for the definition of LineSearchCouples.
     */
    typedef enum
    {
        /// No couples.
        COUPLES_NONE = 0,
        /// Couple linesearch of slack and state variables.
        COUPLE_STATE_SLACK = 1,
        /// Couple linesearch of slack variables, state variables and lagrange multipliers.
        COUPLE_STATE_SLACK_LAMBDA = 2
    } LineSearchCouples;

    /**
     * \brief Flags for the definition of LineSearchCouples.
     */
    typedef enum
    {
        /// No couples.
        LINESEARCH_NONE = 0,
        /// Armijo linesearch with L1-penalty-function.
        LINESEARCH_ARMIJO_L1_PENALTY = 1,
    } LineSearchMethod;

    template <typename real>
    class IPparameter
    {
      public:
        /**
         * \brief IPparameter:
         *        Constructor of the class IPparameter.
        \*/
        IPparameter()
        {
            // Set default values:
            this->mu_0                            = real(0.5);
            this->epsilon_tol                     = real(1e-8);
            this->mu_tol                          = real(1e-8);
            this->Theta_mu                        = real(3.0 / 2.0);
            this->kappa_epsilon                   = real(10.0);
            this->kappa_mu                        = real(0.9);
            this->delta_max                       = real(1e+8);
            this->max_iter                        = 1000;
            this->alpha_max                       = real(1.0);
            this->tol_alpha                       = real(2e-8);
            this->sigma                           = real(0.01);
            this->beta                            = real(0.9);
            this->tau_min                         = real(1e-4);
            this->lambda_tol                      = real(1e-10);
            this->RegDiagMatrixEqu                = real(0);
            this->RegDiagMatrixInitialGuessLambda = real(1e-9);
            this->PrintLev                        = PRINT_MARKED; // PRINT_MORE;
            this->CheckMatricesFor_INF_NAN        = true;
            this->MyLineSearchCouples             = COUPLE_STATE_SLACK;
            this->UseMuSequ                       = false;
            this->FailSaveLagrangeInitialValue    = 1e-4;
            this->BreakpointMaxTime               = false;
            this->maximalTime                     = real(1.0);
        }

        /**
         * \brief ~IPparameter:
         *        Virtual destructor of the class IPparameter.
        \*/
        virtual ~IPparameter()
        {
        }

        /*\
        :|:             _    _ _                   _      _    _
        :|:   _ __ _  _| |__| (_)__  __ ____ _ _ _(_)__ _| |__| |___ ___
        :|:  | '_ \ || | '_ \ | / _| \ V / _` | '_| / _` | '_ \ / -_|_-<
        :|:  | .__/\_,_|_.__/_|_\__|  \_/\__,_|_| |_\__,_|_.__/_\___/__/
        :|:  |_|
        \*/

        /// Startvalue of the barrier parameter \f$\mu\f$.
        real mu_0;
        /// Tolerance of the interior-point-method.
        real epsilon_tol;
        /// Tolerance of the barrier parameter \f$\mu\f$ of the interior-point-method.
        real mu_tol;
        /// Correction parameter of the barrier parameter \f$\mu\f$ (\f$ \Theta^{\mu} \in (1,2) \f$).
        real Theta_mu;
        /// Optimazation constant. \f$ \kappa_{\epsilon} > 0 \f$
        real kappa_epsilon;
        /// Optimazation constant. \f$ \kappa_{\mu} \in (0,1) \f$
        real kappa_mu;
        /// Scaling term.
        real delta_max;
        /// Max number of iterations.
        int_type max_iter;
        /// Max stepsize.
        real alpha_max;
        /// Min stepsize.
        real tol_alpha;
        /// Parameter for the armijo-rule. \f$\sigma \in (0,1)\f$
        real sigma;
        /// Parameter for the armijo-rule. \f$\beta \in (0,1)\f$
        real beta;
        /// Distance parameter for the bounds.
        real tau_min;
        /// Printlevel of IPbasic.
        Printlevel PrintLev;
        /// Tolerance of the lagrange multipliers.
        real lambda_tol;
        /// Regularisation parameter of the diagonalmatrix of the equations.
        real RegDiagMatrixEqu;
        /// Regularisation parameter of the diagonalmatrix lagrange multipliers.
        real RegDiagMatrixInitialGuessLambda;
        /// Tells if matrices are checked due to inf and nan.
        bool CheckMatricesFor_INF_NAN;
        /// Used LineSearchCouples.
        LineSearchCouples MyLineSearchCouples;
        /// Tells, if mu sequence should be used.
        bool UseMuSequ;
        /// Failsave lagrange initial value.
        real FailSaveLagrangeInitialValue;
        /// Tells, if there is a breakpoint with a maximal time.
        bool BreakpointMaxTime;
        /// If BreakpointMaxTime, maximal time for IPbasic to solve.
        real maximalTime;
    };

} // namespace IPbasic

#endif // IPBASIC_PARAMETER_H
