/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file IPbasic_Solver.h
 * Header definitions of the class Solver.
 *
 * @author A. Huber
 * @since 05.02.2019
\*/

#ifndef IPBASIC_SOLVER_H
#define IPBASIC_SOLVER_H

#include "IPbasic.h"
#include "IPbasic_Data.h"
#include "IPbasic_KKTSystem.h"
#include "IPbasic_NLP.h"
#include "IPbasic_Parameter.h"
#include "LinSolvers/IPbasic_LinearSolver_interface.h"
#include "LineSearch/IPbasic_LineSearch.h"
#include "LineSearch/IPbasic_LineSearchFunction.h"

namespace IPbasic
{

    template <typename real>
    class Solver
    {
      public:
        /*!
         * \brief Solver:
         *        Constructor of the class Solver.
         *
         * \param[in] Problem               NLP-Problem, which has to be solved.
         * \param[in] LinSolver             Solver of the linear system.
         * \param[in] IPParams              Interior-point-parameters.
         * \param[in] Timings               Timings of the interior-point-method.
         * \param[in] UsedLineSearchMethod  Used LineSearchMethod.
         *
         */
        Solver(NLP<real> *                    Problem,
               LinearSolver_interface<real> * LinSolver,
               IPparameter<real> *            IPParams             = nullptr,
               IPtimings<real> *              Timings              = nullptr,
               LineSearchMethod               UsedLineSearchMethod = LINESEARCH_ARMIJO_L1_PENALTY);

        /*!
         * \brief ~Solver:
         *        Virtual destructor of the class Solver.
         */
        virtual ~Solver();

        /*\
        :|:             _    _ _                      _
        :|:   _ __ _  _| |__| (_)__   _ __  ___ _ __ | |__  ___ _ _ ___
        :|:  | '_ \ || | '_ \ | / _| | '  \/ -_) '  \| '_ \/ -_) '_(_-<
        :|:  | .__/\_,_|_.__/_|_\__| |_|_|_\___|_|_|_|_.__/\___|_| /__/
        :|:  |_|
        \*/

        /*!
         * \brief checkMaxIter:
         *        True if MaxIter was reached.
         *
         * \return True if MaxIter was reached.
         *
         */
        bool checkMaxIter();

        /*!
         * \brief getIPparameter:
         *        Retruns the parameter of the interior-point-method.
         *
         * \return Retruns the parameter of the interior-point-method.
         *
         */
        IPparameter<real> * getIPparameter();

        /*!
         * \brief getIPtimings:
         *        Returns the IPtimings.
         *
         * \return Returns the IPtimings.
         *
         */
        IPtimings<real> * getIPtimings();

        /*!
         * \brief optimize:
         *        Optimize the nlp.
         *
         * \return True, if the problem could be solved.
         *
         */
        bool optimize();

        /*!
         * \brief copySolution:
         *        Copy the solution of the interior-point-method.
         *
         * \param[out] x    State solution of the interior-point-method.
         *
         */
        void copySolution(real x[]) const;

        /*!
         * \brief copySolution:
         *        Copy the solution of the interior-point-method.
         *
         * \param[out] x        State solution of the interior-point-method.
         * \param[out] slack    Slack solution of the interior-point-method.
         * \param[out] lambda   Lagrange multiplier solution of the interior-point-method.
         *
         */
        void copySolution(real x[], real slack[], real lambda[]) const;

        /*!
         * \brief getLastIterationDataPoint:
         *        Get the last IterationDataPoint.
         *
         * \return Last IterationDataPoint.
         *
         */
        IterationDataPoint<real> getLastIterationDataPoint() const;

      protected:
        /*\
        :|:                _          _          _                _      _    _
        :|:   _ __ _ _ ___| |_ ___ __| |_ ___ __| | __ ____ _ _ _(_)__ _| |__| |___ ___
        :|:  | '_ \ '_/ _ \  _/ -_) _|  _/ -_) _` | \ V / _` | '_| / _` | '_ \ / -_|_-<
        :|:  | .__/_| \___/\__\___\__|\__\___\__,_|  \_/\__,_|_| |_\__,_|_.__/_\___/__/
        :|:  |_|
        \*/

        /// Data of the interior-point-method.
        IPdata<real> * InteriorPointData;

        /// KKT-system of the interior-point-method.
        KKTSystem<real> * IPKKTSystem;

        /// Linesearch method.
        LineSearch<real> * IPLineSearch;

        /// Linesearchfunction.
        LineSearchFunction<real> * IPLineSearchFunction;

        /// Timer to messure the time the optimizer needs.
        TicToc TimerOptimize;

        /// Timer to messure local times.
        TicToc LocalTimerSolver;

        /*\
        :|:                _          _          _                   _
        :|:   _ __ _ _ ___| |_ ___ __| |_ ___ __| |  _ __  ___ _ __ | |__  ___ _ _ ___
        :|:  | '_ \ '_/ _ \  _/ -_) _|  _/ -_) _` | | '  \/ -_) '  \| '_ \/ -_) '_(_-<
        :|:  | .__/_| \___/\__\___\__|\__\___\__,_| |_|_|_\___|_|_|_|_.__/\___|_| /__/
        :|:  |_|
        \*/

        /*!
         * \brief checkForValidMembers:
         *        Checks if the membervariables are set.
         *
         * \return True if the membervariables are set. Else false.
         *
         */
        bool checkForValidMembers();

        /*!
         * \brief checkFinished:
         *        Checkes, if the optimizer is converged.
         *
         * \return True, if the optimizer is converged.
         *
         */
        bool checkFinished();

        /*!
         * \brief getAlphaMax:
         *        Calculates the current maximal stepsizes.
         *
         * \param[out] alphaMax_s       Maximal stepsize.
         * \param[out] alphaMax_lambda  Maximal stepsize for the lagrange multipliers.
         *
         */
        void getAlphaMax(real & alphaMax_s, real & alphaMax_lambda);

        /*!
         * \brief applyStep:
         *        Applies the new step.
         *
         * \param[in] alpha    Stepsize for the states.
         * \param[in] alpha_s  Stepsize for the slack variables.
         * \param[in] alpha_l  Stepsize for the lagrange multipliers.
         *
         */
        void applyStep(real alpha, real alpha_s, real alpha_l);

        /*!
         * \brief updateMu:
         *        Makes an update for the current barrier parameter \f$\mu\f$.
         *
         * \param[in,out] mu        Barrier parameter \f$\mu\f$.
         * \param[in]     Error     Error in the KKT conditions.
         *
         */
        virtual void updateMu(real & mu, real Error);
    };

} // namespace IPbasic

#endif // IPBASIC_SOLVER_H
