/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file IPbasic_LinearSolver_MA57.h
 * Header definitions of the class LinSolver.
 *
 * @author A.Huber and E.Bertolazzi
 * @since 25.03.2019
\*/

#ifndef IPBASIC_LINSOLVER_MA48_H
#define IPBASIC_LINSOLVER_MA48_H

#include "../Errors/IPbasic_Errors.h"
#include "../interfaces/HSL/cppma48.h"
#include "IPbasic_LinearSolver_interface.h"

#ifdef IPBASIC_USE_HSL

namespace IPbasic
{

    template <typename real>
    class LinearSolver_MA48 : public IPbasic::LinearSolver_interface<real>
    {
      private:
        /**
         * \brief checkLastDim:
         *        Checks the last dimensions.
         *
         * \param[in] DimRow    Number of rows.
         * \param[in] DimCol    Number of columns.
         * \param[in] nnz       Number of nonzeros.
         *
         * @return True if the last dimensions are the same.
         *
        \*/
        bool checkLastDim(int_type DimRow, int_type DimCol, int_type nnz)
        {
            if (DimRow != this->last_DimRow)
                return false;
            if (DimCol != this->last_DimCol)
                return false;
            if (nnz != this->last_nnz)
                return false;
            return true;
        }

        /// Definition of HSL solver.
        HSL_IFACE::CPPMA48<real> MyHSL;

        /// Last number of rows.
        int_type last_DimRow;
        /// Last number of columns.
        int_type last_DimCol;
        /// Last number of nonzeros.
        int_type last_nnz;
        /// Tells if symbolic factorisation is done all the time.
        bool doSymbolic;

      public:
        /**
         * \brief LinSolver:
         *        Contrictor of the class LinSolver.
        \*/
        LinearSolver_MA48(bool doSymbolic = false)
            : IPbasic::LinearSolver_interface<real>()
        {
            this->last_DimRow = 0;
            this->last_DimCol = 0;
            this->last_nnz    = 0;
            this->doSymbolic  = doSymbolic;
        }

        /**
         * \brief ~LinSolver:
         *        Virtual destructor of the class LinSolver.
        \*/
        virtual ~LinearSolver_MA48()
        {
        }

        /*\
        :|:   ___      _    _ _      __  __           _
        :|:  | _ \_  _| |__| (_)__  |  \/  |___ _ __ | |__  ___ _ _ ___
        :|:  |  _/ || | '_ \ | / _| | |\/| / -_) '  \| '_ \/ -_) '_(_-<
        :|:  |_|  \_,_|_.__/_|_\__| |_|  |_\___|_|_|_|_.__/\___|_| /__/
        \*/

        /**
         * \brief factorize:
         *        Factorizes the matrix \f$ A\f$ for a sparse matrix.
         *
         * \param[in]     Sparse  Sparse matrix which has to be solved.
         *
         * @return True if the the matrix \f$ A\f$ could be factorized.
         *
         */
        virtual bool factorize(SparseMatrixBase<real> const & Sparse)
        {
            int_type n_row, n_col, nnz;
            Sparse.get_info(n_row, n_col, nnz);
            int_type const * i_Row;
            int_type const * j_Col;
            real const *     val;
            Sparse.get_data(i_Row, j_Col, val);
            bool isFortran = Sparse.FORTRAN_indexing();
            if (!this->MyHSL.checkInitialized() || !this->checkLastDim(n_row, n_col, nnz) || this->doSymbolic)
            {
                if (!this->MyHSL.init(nnz, n_row, n_col, i_Row, j_Col, isFortran))
                {
                    IPbasic::Errors::set_Error("Could not initialize HSL!");
                    return false;
                }
            }
            if (!this->MyHSL.factorize(val))
            {
                IPbasic::Errors::set_Error("Could not factorize HSL!");
                return false;
            }
            return true;
        }

        /**
         * \brief solve:
         *        Solves the linear system \f$ Ax = b\f$ for \f$ A\f$
         *        for a sparse matrix.
         *
         * \param[in]     Sparse  Sparse matrix which has to be solved.
         * \param[in,out] RHS     In:  right-hand-side of the linear system.
         *                        OUT: solution of the linear system.
         *
         * @return True if the linear system \f$ Ax = b\f$ for \f$ A\f$ could be solved.
         *
        \*/
        virtual bool solve(SparseMatrixBase<real> const & Sparse, real RHS[])
        {
            int_type n_row, n_col, nnz;
            Sparse.get_info(n_row, n_col, nnz);
            return this->MyHSL.solve(1, RHS, n_col, RHS, n_col);
        }
    };
} // namespace IPbasic

#else  // not defined IPBASIC_USE_HSL
namespace IPbasic
{

    template <typename real>
    class LinearSolver_MA48 : public IPbasic::LinearSolver_interface<real>
    {
      private:
        /// Tells if symbolic factorisation is done all the time.
        bool doSymbolic;

      public:
        /**
         * \brief LinSolver:
         *        Contrictor of the class LinSolver.
        \*/
        LinearSolver_MA48(bool doSymbolic = false)
            : IPbasic::LinearSolver_interface<real>()
        {
            this->doSymbolic = doSymbolic;
        }

        /**
         * \brief ~LinSolver:
         *        Virtual destructor of the class LinSolver.
        \*/
        virtual ~LinearSolver_MA48()
        {
        }

        /*\
        :|:   ___      _    _ _      __  __           _
        :|:  | _ \_  _| |__| (_)__  |  \/  |___ _ __ | |__  ___ _ _ ___
        :|:  |  _/ || | '_ \ | / _| | |\/| / -_) '  \| '_ \/ -_) '_(_-<
        :|:  |_|  \_,_|_.__/_|_\__| |_|  |_\___|_|_|_|_.__/\___|_| /__/
        \*/

        /**
         * \brief factorize:
         *        Factorizes the matrix \f$ A\f$ for a sparse matrix.
         *
         * \param[in]     Sparse  Sparse matrix which has to be solved.
         *
         * @return True if the the matrix \f$ A\f$ could be factorized.
         *
         */
        virtual bool factorize(SparseMatrixBase<real> const &)
        {
            IPbasic::Errors::set_Error("Can not factorize linear system because HSL is not defined!");
            return false;
        }

        /**
         * \brief solve:
         *        Solves the linear system \f$ Ax = b\f$ for \f$ A\f$
         *        for a sparse matrix.
         *
         * \param[in]     Sparse  Sparse matrix which has to be solved.
         * \param[in,out] RHS     In:  right-hand-side of the linear system.
         *                        OUT: solution of the linear system.
         *
         * @return True if the linear system \f$ Ax = b\f$ for \f$ A\f$ could be solved.
         *
        \*/
        virtual bool solve(SparseMatrixBase<real> const &, real[])
        {
            IPbasic::Errors::set_Error("Can not solve linear system because HSL is not defined!");
            return false;
        }
    };
} // namespace IPbasic
#endif // IPBASIC_USE_HSL

#endif // IPBASIC_LINSOLVER_MA57_H
