/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file IPbasic_LinearSolver_interface.h
 * Header definitions of the class LinearSolver_interface.
 *
 * @author A. Huber and E.bertolazzi
 * @since 04.02.2019
 */

#ifndef IPBASIC_LINSOLVER_IFACE_H
#define IPBASIC_LINSOLVER_IFACE_H

#include "../IPbasic.h"
#include <fstream>

namespace IPbasic
{

    template <typename real>
    class LinearSolver_interface
    {
      protected:
        /*!
         * \brief LinearSolver_interface:
         *        Protected constructor of the class LinearSolver_interface.
         */
        LinearSolver_interface()
        {
        }

      public:
        /*!
         * \brief ~LinearSolver_interface:
         *        Virtual destructor of the class LinearSolver_interface.
         */
        virtual ~LinearSolver_interface()
        {
        }

        /*\
        :|:   ___      _    _ _      __  __           _
        :|:  | _ \_  _| |__| (_)__  |  \/  |___ _ __ | |__  ___ _ _ ___
        :|:  |  _/ || | '_ \ | / _| | |\/| / -_) '  \| '_ \/ -_) '_(_-<
        :|:  |_|  \_,_|_.__/_|_\__| |_|  |_\___|_|_|_|_.__/\___|_| /__/
        \*/

        /*!
         * \brief factorize:
         *        Factorizes the matrix \f$ A\f$ for a sparse matrix.
         *
         * \param[in] Sparse               Sparse matrix which has to be solved.
         * \param[in] is_stored_symmetric  If the flag is `true` only the lower part is stored.
         *
         * \return True if the the matrix \f$ A\f$ could be factorized.
         *
         */
        virtual bool factorize(SparseMatrixBase<real> const & Sparse, bool is_stored_symmetric) IPBASIC_PURE_VIRTUAL;

        /*!
         * \brief solve:
         *        Solves the linear system \f$ Ax = b\f$ for \f$ A\f$
         *        for a sparse matrix.
         *
         * \param[in,out] RHS     In:  right-hand-side of the linear system.
         *                        OUT: solution of the linear system.
         *
         * \return True if the linear system \f$ Ax = b\f$ for \f$ A\f$ could be solved.
         *
         */
        virtual bool solve(real RHS[]) IPBASIC_PURE_VIRTUAL;

        /*!
         * \brief printSystemToFiles:
         *
         * \param[in] Sparse          Sparse matrix which has to be solved.
         * \param[in] RHS             Right-hand-side of the linear system.
         * \param[in] FileNameSystem  File for the matrix system.
         * \param[in] FileNameRHS     File for the right-hand-side and the solution of the system.
         *
         */
        void printSystemToFiles(SparseMatrixBase<real> const & Sparse, bool is_stored_symmetric, real const RHS[], string_type const & FileNameSystem, string_type const & FileNameRHS)
        {
            size_t DimRow = size_t(Sparse.get_number_of_rows());
            size_t DimCol = size_t(Sparse.get_number_of_cols());
            if (DimRow == DimCol)
            {
                std::vector<real> SOL(DimRow);
                std::copy_n(RHS, DimRow, SOL.begin());
                std::ofstream fs;

                // Print system:
                fs.open(FileNameSystem.c_str());
                Sparse.print(fs);
                fs.close();

                // Solve system:
                this->factorize(Sparse, is_stored_symmetric);
                this->solve(&SOL.front());

                // Print vectors:
                fs.open(FileNameRHS.c_str());
                fs << std::scientific;
                for (size_t i = 0; i < DimRow; ++i)
                    fs << SOL[i] << ", " << RHS[i] << '\n';
                fs.close();
            }
        }
    };

} // namespace IPbasic

#endif // IPBASIC_LINSOLVER_IFACE_H
