/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file IPbasic_L1_Penalty.h
 * Header definitions of the class L1_Penalty.
 *
 * @author A. Huber
 * @since 08.02.2019
 */

#ifndef IPBASIC_L1_PENALTY_H
#define IPBASIC_L1_PENALTY_H

#include "IPbasic_LineSearchFunction.h"

namespace IPbasic
{

    template <typename real>
    class L1_Penalty : public LineSearchFunction<real>
    {
      private:
        Malloc<real> mem_real;

        /*\
        :|:            _          _                      _      _    _
        :|:   _ __ _ _(_)_ ____ _| |_ ___  __ ____ _ _ _(_)__ _| |__| |___ ___
        :|:  | '_ \ '_| \ V / _` |  _/ -_) \ V / _` | '_| / _` | '_ \ / -_|_-<
        :|:  | .__/_| |_|\_/\__,_|\__\___|  \_/\__,_|_| |_\__,_|_.__/_\___/__/
        :|:  |_|
        \*/

        /// Local memory for states.
        real * local_states;
        /// Local memory for slack variables.
        real * local_slack;
        /// Local memory for the constraints.
        real * local_constraints;
        /// Penalty parameter vector \f$\eta\f$.
        real * eta;

        /*\
        :|:   ___     _          _         __  __           _
        :|:  | _ \_ _(_)_ ____ _| |_ ___  |  \/  |___ _ __ | |__  ___ _ _ ___
        :|:  |  _/ '_| \ V / _` |  _/ -_) | |\/| / -_) '  \| '_ \/ -_) '_(_-<
        :|:  |_| |_| |_|\_/\__,_|\__\___| |_|  |_\___|_|_|_|_.__/\___|_| /__/
        \*/

        /**
         * \brief LocalConstraintsToNablaD:
         *        Sets the vector to the gradient times d.
         *
         * \param[in] IPD     Data of the interior-point-method.
         *
        \*/
        void LocalConstraintsToNablaD(IPdata<real> * IPD);

        /**
         * \brief updatePenaltyParams:
         *        Update der Penaltyparameter \f$\eta\f$.
         *
         * \param[in] IPD     Data of the interior-point-method.
         * \param[in] alpha   Stepsize.
         *
         */
        void updatePenaltyParams(IPdata<real> * IPD, real alpha);

      public:
        /**
         * \brief L1_Penalty:
         *        Constructor of the class L1_Penalty.
        \*/
        L1_Penalty(IPdata<real> * IPD);

        /**
         * \brief ~L1_Penalty:
         *        Virtual destructor of the class L1_Penalty.
        \*/
        virtual ~L1_Penalty();

        /*\
        :|:             _    _ _                      _
        :|:   _ __ _  _| |__| (_)__   _ __  ___ _ __ | |__  ___ _ _ ___
        :|:  | '_ \ || | '_ \ | / _| | '  \/ -_) '  \| '_ \/ -_) '_(_-<
        :|:  | .__/\_,_|_.__/_|_\__| |_|_|_\___|_|_|_|_.__/\___|_| /__/
        :|:  |_|
        \*/

        /**
         * \brief getF:
         *        Get the value of the line-search-function.
         *
         * \param[in] IPD   Data of the interior-point-method.
         * \param[in] alpha Stepsize.
         *
         * @return The value of the line-search-function.
         *
        \*/
        real getF(IPdata<real> * IPD, real alpha);

        /**
         * \brief getDF:
         *        Get the derivative of the line-search-function.
         *
         * \param[in] IPD   Data of the interior-point-method.
         *
         * @return The derivative of the line-search-function.
         *
        \*/
        real getDF(IPdata<real> * IPD);
    };

} // namespace IPbasic

#endif // IPBASIC_L1_PENALTY_H
