/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file cppma48.h
 * Header definitions of the class CPPMA48.
 *
 * @author A. Huber and E.Bertolazzi
 * @since 15.11.2018
\*/

#include "../../IPbasic_Config.h"

#ifdef IPBASIC_USE_HSL

#ifndef CPPMA48_H
#define CPPMA48_H

#include "hslsolver.h"
#include <vector>

namespace HSL_IFACE
{

    template <typename real>
    class CPPMA48 : public HSLSolver<real>
    {
      private:
        // Memoryvectors for int.
        std::vector<int> i_Row_stored;
        std::vector<int> j_Col_stored;
        std::vector<int> irn;
        std::vector<int> jcn;
        std::vector<int> iw;
        std::vector<int> keep;

        // Memoryvectors for real.
        std::vector<real> a;
        std::vector<real> w;
        std::vector<real> tmpRHS;

        /// Integer, der Aufgabe festlegt (1 für löse AX=B).
        int job;

        /// Länge der Arrays a, irn und jcn (meist 3*DimSparse).
        int la;
        /// Array der Länge 5 mit real Steuerungsparametern.
        real cntl[10];
        /// Array der Länge 20 mit int Steuerungsparametern.
        int icntl[20];
        /// /// Error Array von MA48.
        real ma48errors[3];

        /// Integer Info Variablen.
        int info[20];
        /// Real Info Variablen.
        real rinfo[10];

        void load_error_string(int info);

        // Initialize MA48:
        void ma48i(real cntl[], int icntl[]) const;

        void ma48a(int nrow, int ncol, int nz, int job, int la, real a[], int irn[], int jcn[], int keep[], real const cntl[10], int const icntl[20], int iw[], int info[20], real rinfo[10]) const;

        void
        ma48b(int nrow, int ncol, int nz, int job, int la, real a[], int irn[], int jcn[], int const keep[], real const cntl[10], int const icntl[20], real w[], int iw[], int info[20], real rinfo[10])
            const;

        void ma48c(int        nrow,
                   int        ncol,
                   int        itrans,
                   int        job,
                   int        la,
                   real const a[],
                   int const  irn[],
                   int const  keep[],
                   real const cntl[10],
                   int const  icntl[20],
                   real const rhs[],
                   real       x[],
                   real       errors[3],
                   real const w[],
                   int const  iw[],
                   int        info[20]);

      public:
        /**
         * \brief CPPMA48:
         *        Constructor of the class CPPMA48.
        \*/
        CPPMA48()
        {
        }

        /**
         * \brief ~CPPMA48:
         *        Virtual destructor of the class CPPMA48.
        \*/
        virtual ~CPPMA48()
        {
        }

        /*\
        :|:   ___      _    _ _      __  __           _
        :|:  | _ \_  _| |__| (_)__  |  \/  |___ _ __ | |__  ___ _ _ ___
        :|:  |  _/ || | '_ \ | / _| | |\/| / -_) '  \| '_ \/ -_) '_(_-<
        :|:  |_|  \_,_|_.__/_|_\__| |_|  |_\___|_|_|_|_.__/\___|_| /__/
        \*/

        /**
         * \brief init:
         *        Initializes the HSL solver and does a symbolic solve.
         *
         * \param[in] nnz               Number of non zeros.
         * \param[in] N_Row             Row dimension of the matrix.
         * \param[in] N_Col             Col dimension of the matrix.
         * \param[in] i_Row             Rows of the matrix.
         * \param[in] j_Col             Cols of the matrix.
         * \param[in] isFortranIndexing True for Fortran indexing and false for C indexing.
         *
         * @return True if HSL could be initialized.
         *
        \*/
        virtual bool init(int nnz, int N_Row, int N_Col, int const i_Row[], int const j_Col[], bool isFortranIndexing);

        /**
         * \brief factorize:
         *        Factorizes the linear system \f$ Ax = b\f$ for \f$ A\f$ for a sparse matrix.
         *
         * \param[in] ArrayA Values of the sparse matrix.
         *
         * @return True if the linear system \f$ Ax = b\f$ for \f$ A\f$ could be solved.
         *
        \*/
        virtual bool factorize(real const ArrayA[]);

        bool solve(real const RHS[], real X[], bool itrans);

        virtual bool solve(int nrhs, real const RHS[], int ldRHS, real X[], int ldX);

        virtual bool solve_transposed(int nrhs, real const RHS[], int ldRHS, real X[], int ldX);
    };

} // namespace HSL_IFACE

#endif // CPPMA48_H

#endif // IPBASIC_USE_HSL
