/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file cppma57.h
 * Header definitions of the class CPPMA57.
 *
 * @author A. Huber and E.Bertolazzi
 * @since  13.11.2018
\*/

#include "../../IPbasic_Config.h"

#ifdef IPBASIC_USE_HSL

#ifndef CPPMA57_H
#define CPPMA57_H

#include "hslsolver.h"
#include <cmath>
#include <vector>

namespace HSL_IFACE
{

    template <typename real>
    class CPPMA57 : public HSLSolver<real>
    {
      private:
        // Memoryvectors for int.
        std::vector<int> i_Row;
        std::vector<int> j_Col;
        std::vector<int> iKeep;
        std::vector<int> iPivotSeq;
        std::vector<int> ifact;

        // Memoryvectors for real.
        std::vector<real> a_stored;
        std::vector<real> fact;
        std::vector<real> Work;
        std::vector<real> RHSReinfinement;
        std::vector<real> residualVec;

        // MA57 Speicher:
        /// Array der Länge 5 mit real Steuerungsparametern.
        real cntl[5];
        /// Array der Länge 20 mit int Steuerungsparametern.
        int icntl[20];
        /// Integer Info Variablen.
        int iinfo[40];
        /// Real Info Variablen.
        real rinfo[20];

        // Faktoren in MA57:
        // Workspace MA57:

        /// Integer, der Aufgabe festlegt (1 für löse AX=B).
        int job;

        /// Gibt an, ob iterativ verfeinert werden soll.
        bool doReinfinement;
        /// Residuum tolerance.
        real tolRes;
        /// Maximum number of refinements.
        int MaxRefinements;

        /**
         * \brief getResidual:
         *        Get the maximum from the residual vector.
         *
         * \param[in] _residualVec The residual vector.
         *
         * @return The maximum from the residual vector.
         */
        real getResidual(std::vector<real> const & _residualVec)
        {
            typename std::vector<real>::const_iterator it  = _residualVec.begin();
            real                                       ret = real(0);
            while (it != _residualVec.end())
            {
                real av = std::abs(*it++);
                if (av > ret)
                    ret = av;
            }
            return ret;
        }

        void ma57i(real cntl[], int icntl[]) const;

        void ma57a(int n, int nz, int const irn[], int const jcn[], int lkeep, int keep[], int iw[], int const icntl[20], int info[40], real rinfo[20]) const;

        void ma57b(int        n,
                   int        nz,
                   real const a[],
                   real       fact[],
                   int        lfact,
                   int        ifact[],
                   int        lifact,
                   int        lkeep,
                   int const  keep[],
                   int        iw[],
                   int const  icntl[20],
                   real const cntl[5],
                   int        info[40],
                   real       rinfo[20]) const;

        void ma57c(int job, int n, real const fact[], int lfact, int const ifact[], int lifact, int nrhs, real rhs[], int lrhs, real w[], int lw, int iw[], int const icntl[20], int info[40]) const;

        void ma57d(int        job,
                   int        n,
                   int        ne,
                   real const a[],
                   int const  irn[],
                   int const  jcn[],
                   real const fact[],
                   int        lfact,
                   int const  ifact[],
                   int        lifact,
                   real const rhs[],
                   real       x[],
                   real       resid[],
                   real       w[],
                   int        iw[],
                   int const  icntl[20],
                   real const cntl[5],
                   int        info[40],
                   real       rinfo[20]) const;

      public:
        /**
         * \brief CPPMA57:
         *        Constructor of the class CPPMA57.
        \*/
        CPPMA57()
            : HSLSolver<real>()
        {
            // Default values:
            this->doReinfinement = true;
            this->tolRes         = real(1e-11);
            this->MaxRefinements = 100;
        }

        /**
         * \brief ~CPPMA57:
         *        Virtual destructor of the class CPPMA57.
        \*/
        virtual ~CPPMA57()
        {
        }

        /*\
        :|:   ___      _    _ _      __  __           _
        :|:  | _ \_  _| |__| (_)__  |  \/  |___ _ __ | |__  ___ _ _ ___
        :|:  |  _/ || | '_ \ | / _| | |\/| / -_) '  \| '_ \/ -_) '_(_-<
        :|:  |_|  \_,_|_.__/_|_\__| |_|  |_\___|_|_|_|_.__/\___|_| /__/
        \*/

        /**
         * \brief init:
         *        Initializes the HSL solver and does a symbolic solve.
         *
         * \param[in] nnz               Number of non zeros.
         * \param[in] N_Row             Row dimension of the matrix.
         * \param[in] N_Col             Col dimension of the matrix.
         * \param[in] i_Row             Rows of the matrix.
         * \param[in] j_Col             Cols of the matrix.
         * \param[in] isFortranIndexing True for Fortran indexing and false for C indexing.
         *
         * @return True if HSL could be initialized.
         *
        \*/
        virtual bool init(int nnz, int N_Row, int N_Col, int const i_Row[], int const j_Col[], bool isFortranIndexing);

        /**
         * \brief factorize:
         *        Factorizes the linear system \f$ Ax = b\f$ for \f$ A\f$ for a sparse matrix.
         *
         * \param[in] ArrayA Values of the sparse matrix.
         *
         * @return True if the linear system \f$ Ax = b\f$ for \f$ A\f$ could be solved.
         *
        \*/
        virtual bool factorize(real const ArrayA[]);

        virtual bool solve(int nrhs, real const RHS[], int ldRHS, real X[], int ldX);

        virtual bool solve_transposed(int nrhs, real const RHS[], int ldRHS, real X[], int ldX)
        {
            // Symmetric matrix:
            return this->solve(nrhs, RHS, ldRHS, X, ldX);
        }

        /**
         * \brief setSolverMode:
         *        Switches to a reinfined solver mode.
         *
         * \param[in] DoReinfinement  Tells wether so activate reinfined solver mode.
         * \param[in] TolRes          Residuum tolerance.
         * \param[in] maxRefinements  Maximum number of refinements.
         *
         */
        void setSolverMode(bool DoReinfinement, real TolRes = real(1e-11), int maxRefinements = 100)
        {
            this->doReinfinement = DoReinfinement;
            this->tolRes         = TolRes;
            this->MaxRefinements = maxRefinements;
        }
    };
} // namespace HSL_IFACE

#endif // CPPMA57_H

#endif // IPBASIC_USE_HSL
