/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file hslsolver.h
 * Header definitions of the class HSLSolver.
 *
 * @author A. Huber and E.Bertolazzi
 * @since 13.11.2018
 */

#include "../../IPbasic_Config.h"

#ifdef IPBASIC_USE_HSL

#ifndef HSLSOLVER_H
#define HSLSOLVER_H

#include <string>

// check if compiler is C++11
#if (defined(_MSC_VER) && _MSC_VER >= 1800) || (defined(__cplusplus) && __cplusplus > 199711L)
#define HSL_USE_CXX11
#else
// not C++11 compiler
#ifndef nullptr
#define nullptr NULL
#endif
#endif

#ifndef HSL_USE_CXX11
namespace std
{
    template <class InputIt, class Size, class OutputIt>
    OutputIt copy_n(InputIt first, Size count, OutputIt result)
    {
        return std::copy(first, first + count, result);
    }
} // namespace std
#else
#include <algorithm>
#endif

namespace HSL_IFACE
{

    template <typename real>
    class HSLSolver
    {
      protected:
        int numRows;
        int numCols;
        int nnz;

        /// True if the HSL solver is initialized.
        bool isInitialized;
        /// True if the HSL solver is factorized.
        bool isFactorized;

        std::string last_error;

        /**
         * \brief HSLSolver:
         *        Constructor of the class HSLSolver.
        \*/
        HSLSolver()
            : isInitialized(false)
            , isFactorized(false)
            , last_error("")
        {
        }

      public:
        /**
         * \brief ~HSLSolver:
         *        Virtual destructor of the class HSLSolver.
        \*/
        virtual ~HSLSolver()
        {
        }

        /*\
        :|:   ___      _    _ _      __  __           _
        :|:  | _ \_  _| |__| (_)__  |  \/  |___ _ __ | |__  ___ _ _ ___
        :|:  |  _/ || | '_ \ | / _| | |\/| / -_) '  \| '_ \/ -_) '_(_-<
        :|:  |_|  \_,_|_.__/_|_\__| |_|  |_\___|_|_|_|_.__/\___|_| /__/
        \*/

        /**
         * \brief init:
         *        Initializes the HSL solver and does a symbolic solve.
         *
         * \param[in] nnz               Number of non zeros.
         * \param[in] N_Row             Row dimension of the matrix.
         * \param[in] N_Col             Col dimension of the matrix.
         * \param[in] i_Row             Rows of the matrix.
         * \param[in] j_Col             Cols of the matrix.
         * \param[in] isFortranIndexing True for Fortran indexing and false for C indexing.
         *
         * @return True if HSL could be initialized.
         *
        \*/
        virtual bool init(int nnz, int N_Row, int N_Col, int const i_Row[], int const j_Col[], bool isFortranIndexing) = 0;

        /**
         * \brief factorize:
         *        Factorizes the linear system \f$ Ax = b\f$ for \f$ A\f$ for a sparse matrix.
         *
         * \param[in] ArrayA Values of the sparse matrix.
         *
         * @return True if the linear system \f$ Ax = b\f$ for \f$ A\f$ could be solved.
         *
        \*/
        virtual bool factorize(real const ArrayA[]) = 0;

        /**
         * \brief solve:
         *        Solves the linear system \f$ Ax = b\f$ for \f$ A\f$ for a sparse matrix.
         *
         * \param[in]  nrhs  number of right-hand-side of the linear system.
         * \param[in]  RHS   right-hand-side(s) of the linear system.
         * \param[in]  ldRHS leading dimension of RHS
         * \param[out] X     solution of the linear system.
         * \param[in]  ldX   leading dimension of X
         *
         * @return True if the linear system \f$ Ax = b\f$ for \f$ A\f$ could be solved.
         *
        \*/
        virtual bool solve(int nrhs, real const RHS[], int ldRHS, real X[], int ldX) = 0;

        /**
         * \brief solve_transposed:
         *        Solves the linear system \f$ A^\top x = b\f$ for \f$ A\f$ for a sparse matrix.
         *
         * \param[in]  nrhs  number of right-hand-side of the linear system.
         * \param[in]  RHS   right-hand-side(s) of the linear system.
         * \param[in]  ldRHS leading dimension of RHS
         * \param[out] X     solution of the linear system.
         * \param[in]  ldX   leading dimension of X
         *
         * @return True if the linear system \f$ A^\top x = b\f$ for \f$ A\f$ could be solved.
         *
        \*/
        virtual bool solve_transposed(int nrhs, real const RHS[], int ldRHS, real X[], int ldX) = 0;

        /**
         * \brief checkInitialized:
         *        Checks the flag isInitialized.
         *
         * @return Returns the flag isInitialized.
         *
         */
        bool checkInitialized()
        {
            return this->isInitialized;
        }

        /**
         * \brief checkFactorized:
         *        Checks the flag isFactorized.
         *
         * @return Returns the flag isFactorized.
         *
         */
        bool checkFactorized()
        {
            return this->isFactorized;
        }
    };

} // namespace HSL_IFACE

#endif // HSLSOLVER_H

#endif // IPBASIC_USE_HSL
